
#ifndef COMPILER_IRInstruction_H
#define COMPILER_IRInstruction_H

#include "BasicBlock.h"
#include "ValueRef.h"
#include "Function.h"
#include "Instruction.h"
#include "codegen/AsmBuilder.hpp"

using namespace std;

enum binaryType {ADD, SUB, MUL, DIV, MOD, AND, OR};
enum cmpType {EQ, NE, LT, LE, GT, GE};


class IRInstruction : public Instruction{
public:

    BasicBlock* block; // nullptr表示全局指令
    string debugInfo;
    static int vRegCount;

    explicit IRInstruction(InstType_Enum instType);

    void outPut() override { cout << debugInfo << endl;}

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override {}

    static void setVRegisterCount(int n){
        vRegCount = n;
    }

    static std::string getLastVRegister(int pre = 1){
        return "VREG" + std::to_string(vRegCount - pre);
    }

    static std::string getVRegister(){
        return "VREG" + std::to_string(vRegCount++);
    }

    static void retFunc(AsmBuilder *pBuilder, int offset){
        auto cur_block = pBuilder->getBlock();
        auto ld1 = new LoadMInstruction(cur_block,
                                        LoadMInstruction::LD,
                                        new MachineOperand(MachineOperand::REG, IREGISTER::RA),
                                        new MachineOperand(MachineOperand::IMM, offset + 16 - 8), // framesize - 8
                                        new MachineOperand(MachineOperand::REG, IREGISTER::SP));
        cur_block->insertInst(ld1);
        auto ld2 = new LoadMInstruction(cur_block,
                                        LoadMInstruction::LD,
                                        new MachineOperand(MachineOperand::REG, IREGISTER::FP),
                                        new MachineOperand(MachineOperand::IMM, offset), // framesize
                                        new MachineOperand(MachineOperand::REG, IREGISTER::SP));
        cur_block->insertInst(ld2);
        auto addi = new BinaryMInstruction(cur_block,
                                           BinaryMInstruction::ADDI,
                                           new MachineOperand(MachineOperand::REG, IREGISTER::SP),
                                           new MachineOperand(MachineOperand::REG, IREGISTER::SP),
                                           new MachineOperand(MachineOperand::IMM, offset+16));
        cur_block->insertInst(addi);
        auto jr = new JrMInstruction(cur_block,
                                     JrMInstruction::JR,
                                     new MachineOperand(MachineOperand::REG, IREGISTER::RA));
        cur_block->insertInst(jr);
    }

    static void loadImmIntoReg(AsmBuilder *pBuilder, long value){
        auto cur_block = pBuilder->getBlock();
        auto liImm = new LoadImmInstruction(cur_block,
                                            LoadImmInstruction::LI,
                                            new MachineOperand(MachineOperand::VREG, getVRegister()),
                                            new MachineOperand(MachineOperand::IMM, value));
        pBuilder->getBlock()->insertInst(liImm);
    }

//    将布尔值取反
    static void xoriReg(AsmBuilder *pBuilder, MachineOperand* t){
        auto cur_block = pBuilder->getBlock();
        auto xori = new BinaryMInstruction(cur_block,
                                           BinaryMInstruction::XORI,
                                           t,
                                           t,
                                           new MachineOperand(MachineOperand::IMM, 1));
        pBuilder->getBlock()->insertInst(xori);
    }

//    binary
    static void BinaryInstr(AsmBuilder *pBuilder, BinaryMInstruction::OpType op, string dst, string src1, string src2){
        auto cur_block = pBuilder->getBlock();
        auto bin = new BinaryMInstruction(cur_block,
                                          op,
                                          new MachineOperand(MachineOperand::VREG, dst),
                                          new MachineOperand(MachineOperand::VREG, src1),
                                          new MachineOperand(MachineOperand::VREG, src2));
        pBuilder->getBlock()->insertInst(bin);
    }

//    add register and imm
    static void Addiw(AsmBuilder *pBuilder, BinaryMInstruction::OpType op, string dst, string src1, int imm){
        auto cur_block = pBuilder->getBlock();
        auto bin = new BinaryMInstruction(cur_block,
                                          op,
                                          new MachineOperand(MachineOperand::VREG, dst),
                                          new MachineOperand(MachineOperand::VREG, src1),
                                          new MachineOperand(MachineOperand::IMM, imm));
        pBuilder->getBlock()->insertInst(bin);
    }

//    jump label
    static void JumpLabel(AsmBuilder *pBuilder, const string& label){
        auto br = new JumpMInstruction(pBuilder->getBlock(),
                                       JumpMInstruction::J,
                                       new MachineOperand(MachineOperand::LABEL, label));
        pBuilder->getBlock()->insertInst(br);
    }

    static void loadFromFP(AsmBuilder *pBuilder,LoadMInstruction::OpType optype, string dst, int imm){
        auto lw1 = new LoadMInstruction(pBuilder->getBlock(),
                                        optype,
                                        new MachineOperand(MachineOperand::VREG, dst),
                                        new MachineOperand(MachineOperand::IMM, imm),
                                        new MachineOperand(MachineOperand::REG, IREGISTER::FP));
        pBuilder->getBlock()->insertInst(lw1);
    }

//    arr
    static void arrInit(std::vector<std::any> values, ArrayType* arrayType, std::stringstream &ss){
        if(values.at(0).type() != typeid(vector<any>)){
            int sizeTotal = arrayType->size;
// one dimension
            for(auto & value : values){
                ss << "\t.word\t" << any_cast<ValueRef*>(value)->get_Ref() << endl;
                sizeTotal -= 4;
            }

            if(sizeTotal > 0)
                ss << "\t.zero\t" << sizeTotal << '\n';
        }
        else{
            for(auto & value : values){
                arrInit(any_cast<vector<any>>(value), dynamic_cast<ArrayType*>(arrayType->elementType) ,ss);
            }
        }
    }

};

class DummyInstruction : public IRInstruction {//NOP指令
public:

    DummyInstruction();

    //void outPut() override;
};

class AddGlobalInstruction : public IRInstruction {
public:
    ValueRef* dst;
    Type* varType;
    ValueRef* initVal;

    AddGlobalInstruction(ValueRef* dst, Type* varType, ValueRef* initVal);

    void codegen(AsmBuilder *pBuilder) const;

    //void outPut() override;
};

class AllocaInstruction : public IRInstruction {
public:
    ValueRef* dst;
    Type* varType;

    AllocaInstruction(ValueRef* dst, Type* varType);

    //void outPut() override;
};

class LoadInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;

    LoadInstruction(ValueRef* dst,ValueRef* src);//(ValueRef* dst, Pointer* src)?

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;
    //void outPut() override;
};

class StoreInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;

    StoreInstruction(ValueRef* dst,ValueRef* src);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;
};
class BinaryInstruction : public IRInstruction {
public:

    ValueRef* dst;
    ValueRef* src1;
    ValueRef* src2;
    binaryType opTy;

    BinaryInstruction(ValueRef* dst,ValueRef* src1, ValueRef* src2, binaryType opTy);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    void outPut() override;
};

class CmpInstruction : public IRInstruction {
public:
    ValueRef* result;
    ValueRef* src1;
    ValueRef* src2;
    cmpType opTy;

    CmpInstruction(ValueRef* result, ValueRef* src1, ValueRef* src2, cmpType opTy);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    void outPut() override;
};

class BrInstruction : public IRInstruction {
public:
    BasicBlock* label;

    BrInstruction(BasicBlock* label);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;

};

class CondBrInstruction : public IRInstruction {
public:
    ValueRef* condition;
    BasicBlock* trueLabel;
    BasicBlock* falseLabel;

    CondBrInstruction(ValueRef* condition, BasicBlock* trueLabel, BasicBlock* falseLabel);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;
};

class RetInstruction : public IRInstruction {
public:
    ValueRef* retVal; // nullptr 表示 ret void;

    RetInstruction(ValueRef* retVal);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;
    //void outPut() override;
};

class CallInstruction : public IRInstruction {
public:
    ValueRef* retVal;
    Function* function;
    vector<ValueRef*> args;

    CallInstruction(ValueRef* retVal, Function* function, vector<ValueRef*>& args);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;
};

class ZExtInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;

    ZExtInstruction(ValueRef* dst,ValueRef* src);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;
};

class XorInstruction : public IRInstruction {
public:
    // 理论上只有Xor true的情况，所以另一个参数永远为1(i1/i32)
    ValueRef* dst;
    ValueRef* src;

    XorInstruction(ValueRef* dst, ValueRef* src);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;


    //void outPut() override;
};

class GEPInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;
    ValueRef* index; // 暂时做成只取一层下标的

    GEPInstruction(ValueRef* dst, ValueRef* src, ValueRef* index);

    void codegen(AsmBuilder *pBuilder,
                 std::map<std::string, int> &offset_table,
                 std::map<std::string, int> &size_table,
                 int frameSize) override;

    //void outPut() override;
};

class PhiInstruction : public IRInstruction {
public:
    ValueRef* symbol; // you shouldn't use this
    ValueRef* result;
    int branch_cnt;
    map<BasicBlock*, ValueRef*> mp;

    PhiInstruction(ValueRef* symbol,ValueRef* result,int cnt);

    void outPut() override;
};

class FPSIInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;

    FPSIInstruction(ValueRef* dst, ValueRef* src);

    //void outPut() override;
};

class FNegInstruction : public IRInstruction {
public:
    ValueRef* dst;
    ValueRef* src;

    FNegInstruction(ValueRef* dst, ValueRef* src);

    //void outPut() override;
};

//class BitCastInstruction : public IRInstruction {//?
//public:
//    enum opType {};
//
//    BitCastInstruction(BasicBlock* block);
//
//    //void outPut() override;
//};
//
//class ShlInstruction : public IRInstruction {//?
//public:
//    enum opType {};
//
//    ShlInstruction(BasicBlock* block);
//
//    //void outPut() override;
//};
//
//class AShrInstruction : public IRInstruction {//?
//public:
//    enum opType {};
//
//    AShrInstruction(BasicBlock* block);
//
//    //void outPut() override;
//};

#endif //COMPILER_IRInstruction_H

